import numpy as np
import pandas as pd
from scipy.interpolate import LinearNDInterpolator
from radpy.config import rppath, rapath, kppath, kapath, hppath, hapath, jppath, japath

def create_ldc_functions(df):
    ############################################################################
    # Function: create_ldc_function                                            #
    # Inputs: df -> the Claret data tables                                     #
    # Outputs: ldc_func -> the function to interpolate the ldc for the filter  #
    # What it does:                                                            #
    #      1. Drops any nans in the table                                      #
    #      2. If the model is the Phoenix models, the interpolator only takes  #
    #         the teff and logg and returns the mus.                           #
    #      3. If the model is the Atlas models, the interpolator takes the     #
    #         teff, logg, and metallicity, and returns the mus.                #
    #      4. Returns the interpolated function                                #
    ############################################################################
    df_clean = df.dropna()
    if df_clean['Mod'][0] == 'P':
        ldc_func = LinearNDInterpolator((df_clean['Teff'], df_clean['logg']), df_clean['u'])
        return ldc_func
    else:
        points = np.stack([df_clean['Teff'], df_clean['logg'], df_clean['Z']], axis = -1)
        ldc_func = LinearNDInterpolator(points, df_clean['u'])
        return ldc_func


def ldc_calc(teff, logg, feh, filt, verbose = False):
    ###########################################################################
    # Function: ldc_calc                                                      #
    # Inputs: teff -> the effective temperature                               #
    #         logg -> the surface gravity                                     #
    #         feh -> the metallicity                                          #
    #         filt -> the filter for the LDC                                  #
    #         verbose -> if set to true, prints the print statements          #
    #                    defaults to not print                                #
    # Outputs: the limb-darkening coefficient                                 #
    # What it does:                                                           #
    #       1. returns the lower-case value for the filter to make sure there #
    #          is no case mismatch                                            #
    #       2. defines a dictionary for the Phoenix functions                 #
    #          Phoenix is used if the temps are below 3500 K                  #
    #       3. defines a dictionary for the Atlas functions                   #
    #          Atlas is used if the temps are above 3500 K                    #
    #       4. if temps are below 3500 and the logg is between 3.5 and 5,     #
    #          uses the Phoenix models to generate a limb darkening coeff.    #
    #       5. if temps are above 3500, uses the Atlas models to generate     #
    #          a limb darkening coeff.                                        #
    #       6. Returns the limb darkening coefficient.                        #
    ###########################################################################
    filt = filt.lower()
    # Functions for teff < 3500
    low_teff_funcs = {
        'r': RPfunc,
        'h': HPfunc,
        'k': KPfunc,
        'j': JPfunc
    }
    # Functions for teff >= 3500
    high_teff_funcs = {
        'r': RAfunc,
        'h': HAfunc,
        'k': KAfunc,
        'j': JAfunc
    }
    if teff < 3500 and 3.5 <=logg <=5:
        if verbose:
            print('Using the Phoenix models.')
        func = low_teff_funcs.get(filt)
        if func is None:
            raise ValueError(f"Unknown filter: {filt}")
        mu = func([teff, logg])[0]
    else:
        if verbose:
            print("Using the Atlas models.")
        func = high_teff_funcs.get(filt)
        if func is None:
            raise ValueError(f"Unknown filter: {filt}")
        mu = func([teff, logg, feh])[0]
    return mu

ldc_RP = pd.read_csv(rppath)
ldc_RA = pd.read_csv(rapath)
ldc_KP = pd.read_csv(kppath)
ldc_KA = pd.read_csv(kapath)
ldc_HP = pd.read_csv(hppath)
ldc_HA = pd.read_csv(hapath)
ldc_JP = pd.read_csv(jppath)
ldc_JA = pd.read_csv(japath)

RPfunc = create_ldc_functions(ldc_RP)
RAfunc = create_ldc_functions(ldc_RA)
HPfunc = create_ldc_functions(ldc_HP)
HAfunc = create_ldc_functions(ldc_HA)
KPfunc = create_ldc_functions(ldc_KP)
KAfunc = create_ldc_functions(ldc_KA)
JPfunc = create_ldc_functions(ldc_JP)
JAfunc = create_ldc_functions(ldc_JA)


import numpy as np
import pandas as pd
from lmfit import Model
import scipy.special as ss
from astropy.stats import mad_std
from radpy.stellar import temp
import warnings
warnings.filterwarnings("ignore", message="Using UFloat objects with std_dev==0 may give unexpected results.")
warnings.filterwarnings("ignore", message="DataFrameGroupBy.apply operated on the grouping columns")
#Uniform disk V2 equation
def UDV2(sf, theta):
    x = np.pi*sf*(theta/(206265*1000))
    vis = (2*ss.jv(1,x))/x
    return vis**2

##########################################################################################
# calculates chi squared and chi squared reduced
def chis(y, exp_y, yerr, dof):
    chisqr = np.sum(((y - exp_y) / (yerr)) ** 2)
    chisqr_red = chisqr / (len(y) - dof)
    return chisqr, chisqr_red


##########################################################################################
# Weighted average function
def weight_avg(df):
    x = df['V2']
    dx = df['dV2']
    weight = 1 / (dx ** 2)
    wavg = sum(weight * x) / sum(weight)
    dwavg = 1 / np.sqrt(sum(weight))
    return dwavg


##########################################################################################
# Random bracket function for bootstrapping
def random_bracket(df, num_of_brackets):
    ###########################################################################
    # Function: random_bracket                                                #
    # Inputs: df -> dataframe with spatial frequency, v2, v2_err, and bracket #
    #         num_of_brackets -> number of brackets in total you have         #
    # Outputs: spf_br -> the spatial frequencies randomized                   #
    #          v2_br -> the visibility squared randomized                     #
    #          dv2_br -> the error on the v2 randomized                       #
    #          wavgs -> weighted averages of the v2                           #
    # What it does:                                                           #
    #      1. sets the seed                                                   #
    #      2. picks a random number between 2 and the number of brackets      #
    #      3. Selects that many unique bracket labels at random               #
    #      4. Filters the dataframe down to only those with those specific    #
    #         bracket labels                                                  #
    #      5. Groups by the bracket labels                                    #
    #      6. Applies the weight average to the grouped data                  #
    #      7. Adds the weighted average as a column in the data frame         #
    #      8. Merges the groups into a new group for weighted average         #
    #         based on the bracket                                            #
    #      9. Splits up the dataframe into spatial frequency, v2, dv2, and    #
    #         wavg                                                            #
    #     10. Returns spatial frequency, v2, dv2, and wavg                    #
    ###########################################################################
    np.random.seed()
    xdata = []
    ydata = []
    dydata = []
    numbr = np.random.randint(2, num_of_brackets)
    # chatgpt wrote the next couple lines
    random_group_ids = df['Bracket'].drop_duplicates().sample(n=numbr).values

    # Filter the DataFrame for the selected groups
    random_groups = df[df['Bracket'].isin(random_group_ids)]

    grouped = random_groups.groupby('Bracket')

    # results = grouped.apply(weight_avg, include_group = False).reset_index()
    results = grouped.apply(weight_avg).reset_index()
    # Version-aware apply for future compatibility
    # pandas_major = int(pd.__version__.split('.')[0])
    # pandas_minor = int(pd.__version__.split('.')[1])
    # if pandas_major >= 2:
    #    results = grouped.apply(weight_avg, include_group=False).reset_index()
    # else:
    #    results = grouped.apply(weight_avg).reset_index()

    results.columns = ['Bracket', 'Wavg']
    random_groups_with_avg = random_groups.merge(results, on='Bracket')

    spf_br = random_groups_with_avg['Spf']  # spatial frequency in rad^-1
    v2_br = random_groups_with_avg['V2']  # Visibility squared
    dv2_br = random_groups_with_avg['dV2']
    wavgs = random_groups_with_avg['Wavg']

    return spf_br, v2_br, dv2_br, wavgs


##########################################################################################
# Percent difference function
def percent_diff(x1, x2, verbose=False):
    diff = (abs(x1 - x2) / ((x1 + x2) / 2)) * 100
    if verbose:
        print("Percent difference:", round(diff, 2), "%")
    return round(diff, 3)

def safe_theta_extraction(result):
    param = result.params['theta']
    value = param.value
    stderr = param.stderr
    if stderr is None or stderr == 0:
        return value, None
    else:
        uvar = result.uvars['theta']
        return uvar.n, uvar.s
##########################################################################################
def initial_UDfit(spf, v2, dv2, theta_guess, star_params, verbose=False):
    #####################################################################
    # Function: initial_UDfit                                           #
    # Inputs: spf -> spatial frequency                                  #
    #         v2 -> visibilitity squared                                #
    #         dv2 -> error on the V2                                    #
    #         theta_guess -> initial guess for theta                    #
    #         star_params -> stellar class object                       #
    #         verbose -> if set to True, allows print statements        #
    #                    defaults to False                              #
    # Outputs: theta_ilm -> initial uniform disk diameter               #
    #          dtheta_ilm -> error on the diameter                      #
    #          chisqr_ilm -> chi squared reduced value                  #
    # What it does:                                                     #
    #        1. Initialized the model                                   #
    #        2. initializes the parameters                              #
    #        3. Fits for the UD diameter using lmfit                    #
    #           uses for the weights as 1/dv2                           #
    #        4. pulls out the theta, dtheta, and chi squared reduced    #
    #        5. updates the stellar object                              #
    #        6. Returns the theta, dtheta, and chi squared reduced      #
    #####################################################################

    udmodel = Model(UDV2)
    udparams = udmodel.make_params(theta=theta_guess)
    ud_result = udmodel.fit(v2, udparams, sf=spf, weights=1 / (dv2), scale_covar=False)
    theta_ilm, dtheta_ilm = safe_theta_extraction(ud_result)

    chisqr_ilm = ud_result.redchi  # chi squared reduced of the fit
    if verbose:
        print('Initial fit with lmfit:')
        print(ud_result.fit_report())

    star_params.update(udthetai=round(theta_ilm,5), udthetai_err=round(dtheta_ilm,5))
    return theta_ilm, dtheta_ilm, chisqr_ilm

def make_df(B, v2, dv2, wave, band, brack, inst, fulldf=False):
    rand_wave = np.random.normal(wave, band)
    new_spf = B / rand_wave
    if not fulldf:
        return new_spf
    if fulldf:
        new_df = pd.DataFrame(np.column_stack((new_spf, v2, dv2, brack, inst)),
                              columns=['Spf', 'V2', 'dV2', 'Bracket', 'Instrument'])
        return new_df

def bootstrap(df, inst):
    ###########################################################
    # Function: bootstrap                                     #
    # Inputs: df -> the data dataframe                        #
    #         inst -> the intstrument                         #
    # Outputs: the new_df                                     #
    # What it does:                                           #
    #       1. If the instrument is set to c (Classic),       #
    #          samples the V2 on a normal distribution        #
    #          and creates a new dataframe with that.         #
    #       2. If the instrument is any others, determines    #
    #          the number of brackets in the dataset.         #
    #       3. calls the random_bracket function              #
    #       4. samples the V2 on a normal distribution        #
    #       5. creates a new dataframe with that              #
    #       6. Returns the new dataframe                      #
    ###########################################################
    if inst == 'c' or inst == 'C':
        newv2 = np.random.normal(df['V2'], df['dV2'])
        new_df = pd.DataFrame(np.column_stack((df['Spf'], newv2, df['dV2'])), columns=['Spf', 'V2', 'dV2'])
        return new_df
    else:
        num_brackets = df['Bracket'].max()
        spfbr, v2br, dv2br, avgdv2 = random_bracket(df, num_brackets)
        newv2 = np.random.normal(v2br, avgdv2)
        new_df = pd.DataFrame(np.column_stack((spfbr, newv2, dv2br)), columns=['Spf', 'V2', 'dV2'])
        return new_df


def udfit(df, stellar_params, verbose=False):
    #####################################################################
    # Function: udfit                                                   #
    # Inputs: df -> dataframe with data in it                           #
    #         star_params -> stellar class object                       #
    #         verbose -> if set to True, allows print statements        #
    #                    defaults to False                              #
    # Outputs: theta_ud -> initial uniform disk diameter                #
    # What it does:                                                     #
    #        1. Initialized the model                                   #
    #        2. initializes the parameters                              #
    #        3. Fits for the UD diameter using lmfit                    #
    #           uses for the weights as 1/dv2                           #
    #        4. pulls out the theta                                     #
    #        5. Returns the theta                                       #
    #####################################################################
    udmodel = Model(UDV2)
    ud_params = udmodel.make_params(theta=stellar_params.udthetai)
    ud_result = udmodel.fit(df['V2'], ud_params, sf=df['Spf'], weights=1 / (df['dV2']), scale_covar=True)

    theta_ud = ud_result.uvars['theta'].n

    return theta_ud


def run_UDfit(mc_num, bs_num, datasets, stellar_params, verbose=False):
    ######################################################################
    # Function: run_udmcbs_fit                                           #
    # Inputs: mc_num -> number of Monte Carlo iterations                 #
    #         bs_num -> number of bootstrap iterations                   #
    #         datasets -> the datasets you want fit                      #
    #                     format: [inst1, inst2, inst3]                  #
    #         stellar_params -> star object                              #
    #         verbose -> if True, allows print statements                #
    #                    default is False                                #
    # Outputs: UD -> a list of all the uniform disk diameters calculated #
    #                during the fitting routine                          #
    # What it does:                                                      #
    #      1. Initializes the empty list for the diameters               #
    #      2. If verbose is set, it initializes the empty lists for      #
    #         spf, v2, and dv2                                           #
    #      3. enters the Monte Carlo loop                                #
    #      4. Creates the dataframes for each dataset                    #
    #      5. For each dataframe, samples the wavelength of observation  #
    #         on a normal distribution. Then calculates new spatial      #
    #         frequencies                                                #
    #      6. Enters the bootstrapping loop                              #
    #      7. Pulls the instrument for each data set                     #
    #      8. Calls the bootstrap function with the instrument's df and  #
    #         the instrument ID                                          #
    #      9. appends the output of the bootstrap function to a list.    #
    #     10. Concatenates all the dataframes created in the bootstrap   #
    #         loop.                                                      #
    #     11. Calculates the uniform disk diameter with udfit            #
    #     12. Appends the diameter to the list called UD.                #
    #     13. After the loops, returns the UD.                           #
    ######################################################################
    UD = []
    if verbose:
        udmcbs_spf = []
        udmcbs_v2 = []
        udmcbs_dv2 = []

    for _ in range(mc_num):
        dfs = [d.make_df() for d in datasets]

        for df in dfs:
            df['Spf'] = df['B'] / np.random.normal(df['Wave'], df['Band'])

        for _ in range(bs_num):
            bs_dfs = []

            for df in dfs:
                inst = df["Instrument"].iloc[0]
                # print('Instrument:', inst)
                boot_df = bootstrap(df, inst)
                bs_dfs.append(boot_df)

            new_df = pd.concat(bs_dfs, ignore_index=True)

            theta_udbs = udfit(new_df, stellar_params, verbose)

            UD.append(theta_udbs)

            if verbose:
                udmcbs_spf.append(new_df['Spf'])
                udmcbs_v2.append(new_df['V2'])
                udmcbs_dv2.append(new_df['dV2'])
                return UD, udmcbs_spf, udmcbs_v2, udmcbs_dv2

    return UD


def udfit_values(x, y, dy, UD, stellar_params, verbose=False):
    ################################################################
    # Function: udfit_values                                       #
    # Inputs: x -> the spatial frequencies                         #
    #         y -> the V2                                          #
    #        dy -> the error on the V2                             #
    #        UD -> the list of diameters                           #
    #        stellar_params -> the star object                     #
    #        verbose - > if true, returns print statements         #
    # Outputs: None                                                #
    # What it does:                                                #
    #     1. Takes the mean of the uniform disk diameter list      #
    #     2. Takes the median absolute deviation of the UDs        #
    #     3. Calculates the chi squared and chi squared reduced    #
    #        of the uniform disk diameter fit.                     #
    #     4. Calculates the effective temperature using the mean   #
    #     5. Updates the stellar object with the new parameters    #
    ################################################################
    avg_UD = np.mean(UD)
    std_UD = mad_std(UD)

    chisq, chisqr = chis(y, UDV2(x, avg_UD), y, 1)
    teff_ud = temp(stellar_params.fbol, stellar_params.fbol_err, avg_UD, std_UD)

    stellar_params.update(teff=round(teff_ud[0],5), teff_err=round(teff_ud[1],5), udtheta=round(avg_UD,5), udtheta_err=round(std_UD,5))
    if verbose:
        print('Uniform Disk Diameter after MC/BS:', round(avg_UD, 4), '+/-', round(std_UD, 5), 'mas')
        print("Chi-squared:", round(chisq, 3))
        print("Chi-squared reduced:", round(chisqr,3))
        print("Temperature:", round(teff_ud[0], 1), "+/-", round(teff_ud[1], 1), "K")



import numpy as np
from zero_point import zpt
from astroquery.gaia import Gaia
from astroquery.simbad import Simbad
from astroquery.vizier import Vizier


# function to calculate temperature from bolometric flux and angular diameter
def temp(Fbol, dF, theta, dtheta, verbose=False):
    T = 2341 * (Fbol / theta ** 2) ** (1 / 4)
    dT = np.sqrt(
        (2341 / (4 * Fbol ** (3 / 4) * theta ** (1 / 2)) * dF) ** 2
        + (2341 * Fbol ** (1 / 4) / (2 * theta ** (3 / 2)) * dtheta) ** 2
    )
    # dT = T*np.sqrt(((dF/Fbol)*(1/4))**2 + ((dtheta/theta)*(1/2))**2)
    if verbose:
        print("Temperature: ", T, "+/-", dT, "[K]")
    return T, dT


##########################################################################################
# calculates the distance given parallax
def dist_calc(p, p_err, zpc, verbose=False):
    pc = (p + zpc) / 1000
    D = 1 / pc
    dD = D * ((p_err / 1000) / pc)
    if verbose:
        print("Corrected parallax:", round(pc,5) * 1000)
        print("Distance:", round(D,5), "+/-", round(dD,5), "[pc]")

    return D, dD


# Luminosity function
def luminosity(D, dD, Fbol, dF):
    pc_conv = 3.091e16  # meters in a parsec
    d = D * pc_conv * 100
    L = (4 * np.pi * (d ** 2) * Fbol) / (3.846e33)
    dL = L * np.sqrt(((2 * dD) / D) ** 2 + (dF / Fbol) ** 2)
    # return(print("Luminosity: ", L ,"+/-",dL, "[L_solar]"))
    return L, dL


##########################################################################################
# Physical radius function
def ang_to_lin(theta, dtheta, D, dD, verbose=False):
    Rs = 6.957e8  # meters in a solar radii
    pc_conv = 3.091e16  # meters in a parsec
    mas_conv = 206265 * 1000  # radians to mas conversion

    R = (theta * (D * pc_conv)) / (2 * mas_conv * Rs)
    # dR = R*np.sqrt((dtheta/theta)**2 + (dD/D)**2)                         #my equation same as Jonas just different derivation
    dR = (pc_conv / (2 * mas_conv * Rs)) * np.sqrt(
        (dtheta * D) ** 2 + (dD * theta) ** 2)  # jonas equation matches the first one
    if verbose:
        print("Linear Radius: ", R, "+/-", dR, "[R_solar]")
    return R, dR


def gaia_correct_distance(gaiadrname, plx=None, dplx=None, verbose=False):
    #####################################################################
    # Function name: gaia_correct_distance                              #
    # Inputs: gaiadrname -> Gaia ID for the star                        #
    #         plx, dplx -> parallax and error, default is None          #
    #                      if None, function pulls parallax from        #
    #                      Gaia catalog                                 #
    #         verbose flag -> default is False                          #
    #                         if True, allows print statements          #
    # Outputs: the Gaia zero-point corrected distance in parsecs        #
    # What it does:                                                     #
    #    1. reads in the source_id, parallax, and parallax error        #
    #    2. checks to make sure the Gaia name is a string               #
    #    3. splits the Gaia name into just the source id numbers        #
    #    4. checks to see which Gaia DR we need to use                  #
    #    5. queries the correct catalog for the needed parameters       #
    #       source id, ra, dec, phot_g_mean_mag, parallax,              #
    #       parallax error, nu_eff_used_in_astrometry,                  #
    #       pseudocolour, ecliptic latitude, and                        #
    #       astrometric_params_solved                                   #
    #    6. if no parallax/parallax error is given by the user,         #
    #       it sets the parallax pulled from the Gaia DR                #
    #    7. Sets the variable names                                     #
    #    8. Checks to make sure the zero_point function is only         #
    #       interpolating, not extrapolating                            #
    #    9. Calculates the zero-point correction value                  #
    #   10. Calculates the distance                                     #
    #####################################################################
    zpt.load_tables()
    name = check_if_string(gaiadrname, verbose)
    source_id = name.split()[-1]
    if verbose:
        print('Source ID', source_id)

    if 'Gaia DR3' in name:
        job = Gaia.launch_job(f"""
            SELECT source_id, ra, dec, phot_g_mean_mag, parallax, parallax_error, nu_eff_used_in_astrometry, pseudocolour, ecl_lat, astrometric_params_solved
            FROM gaiadr3.gaia_source
            WHERE source_id = {source_id}
        """)
        result = job.get_results()
    elif 'Gaia DR2' in name:
        job = Gaia.launch_job(f"""
            SELECT source_id, ra, dec, phot_g_mean_mag, parallax, parallax_error, nu_eff_used_in_astrometry, pseudocolour, ecl_lat, astrometric_params_solved
            FROM gaiadr2.gaia_source
            WHERE source_id = {source_id}
        """)
        result = job.get_results()
    elif 'Gaia DR1' in name:
        job = Gaia.launch_job(f"""
            SELECT source_id, ra, dec, phot_g_mean_mag, parallax, parallax_error, nu_eff_used_in_astrometry, pseudocolour, ecl_lat, astrometric_params_solved
            FROM gaiadr3.gaia_source
            WHERE source_id = {source_id}
        """)
        result = job.get_results()
    else:
        if verbose:
            print('Incorrect input. Please make sure you have a Gaia ID for DR3, DR2, or DR1.')

    if plx is None:
        plx = result['parallax'][0]
    if dplx is None:
        dplx = result['parallax_error'][0]
    gmag = result['phot_g_mean_mag'][0]
    nueff = result['nu_eff_used_in_astrometry'][0]
    pseudo = result['pseudocolour'][0]
    eclat = result['ecl_lat'][0]
    soltype = result['astrometric_params_solved'][0]

    valid = soltype > 3
    zpcorrect = zpt.get_zpt(gmag[valid], nueff[valid], pseudo[valid], eclat[valid], soltype[valid], _warnings=verbose)[
        0]

    distance = dist_calc(plx, dplx, zpcorrect, verbose)
    d = distance[0]
    dd = distance[1]

    return round(d,5), round(dd,5)


def use_hipparcos(star_name, plx=None, dplx=None, verbose=False):
    ###################################################################
    # Function: use_hipparcos                                         #
    # Inputs: star_name -> name of star                               #
    #         plx, dplx -> parallax and error                         #
    #                      If no user input, pulls value from catalog #
    #         verbose flag -> Default is False                        #
    #                         if True, allows print statements        #
    # Outputs: distance and distance error                            #
    # What it does                                                    #
    #     1. Queries "Hipparcos: the New Reduction" catalog           #
    #     2. if it finds results,                                     #
    #           a. assigns parallax and error if set to None          #
    #           b. calculates distance                                #
    #           c. returns distance                                   #
    #     3. if it doesn't find results:                              #
    #           a. returns None, None                                 #
    ###################################################################

    hip_result = Vizier.query_object(star_name, catalog="I/311/hip2")
    if hip_result:
        if verbose:
            print('Found in Hipparcos catalog')
        hip = hip_result[0]
        if plx is None:
            plx = hip['Plx'][0]
        if dplx is None:
            dplx = hip['e_Plx'][0]
        distance = dist_calc(plx, dplx, 0, verbose)
        d = distance[0]
        dd = distance[1]
        return round(d,5), round(dd,5)
    if not hip_result:
        if verbose:
            print('Not found in Hipparcos catalog.')
        return None, None


def check_if_string(name, verbose=False):
    # checks if the name is a string or not
    # if it isn't, converts to a string
    if not isinstance(name, str):
        if verbose:
            print('Star name is not a valid type. Should be a string. Correcting it now.')
        name = str(name)
        return name
    else:
        return name


def distances(star_name, plx=None, dplx=None, use_Hipp=False, verbose=False):
    ################################################################
    # Function: distances                                          #
    # Inputs: star_name -> star name, needs to be a string         #
    #         plx, dplx -> parallax and error, default is None     #
    #                      if default is set, code will pull the   #
    #                      Gaia parallax or the Hipparcos parallax #
    #         use_Hipp -> indicates if user wants to use Hipparcos #
    #                     instead of Gaia                          #
    #                     default is Gaia                          #
    #         verbose flag -> Default is False                     #
    #                         if True, allows print statements     #
    # Outputs: distance and distance error in units of parsecs     #
    # What it does:                                                #
    #     1. Checks to make sure the star name is a string.        #
    #     If use_Hipp flag is set:                                 #
    #     2. calculates the distance using use_hipparcos           #
    #     If use_Hipp flag is not set or it failed:                #
    #     2. Queries Simbad for a list of star IDs using the name. #
    #     3. Checks which Gaia ID it found. First choice is DR3.   #
    #        If there is a DR3 ID:                                 #
    #        a. pulls the the Gaia DR3 ID.                         #
    #        b. Prints a congrats there was one.                   #
    #        c. Calls the gaia_correct_distance function which     #
    #            returns the gaia zero-point corrected distance    #
    #            and error.                                        #
    #        If there is no DR3 ID:                                #
    #        a. Checks for a Gaia DR2 ID                           #
    #        b. pulls the DR2 and prints a congrats we found one.  #
    #        c. Calls the gaia_correct_distance function which     #
    #           returns the gaia zero-point corrected distance     #
    #           and error.                                         #
    #        If there is no DR2 ID:                                #
    #        a. Checks for a Gaia DR1 ID                           #
    #        b. pulls the DR1 and prints a congrats we found one.  #
    #        c. Calls the gaia_correct_distance function which     #
    #           returns the gaia zero-point corrected distance     #
    #           and error.                                         #
    #        If there is no DR1 ID:                                #
    #        a. Prints "No Gaia ID. Searching for Hipparcos'       #
    #        b. Queries the van leeuwen (2007) updated reduction   #
    #           Hipparcos catalog                                  #
    #        c. Pulls the parallax and parallax error              #
    #        d. Calculates the distance                            #
    #     4. Returns the distance.                                 #
    ################################################################
    star_name = check_if_string(star_name, verbose)
    used_hipparcos = False

    if use_Hipp:
        if verbose:
            print('Using Hipparcos parallaxes instead of Gaia.')
        d, dd = use_hipparcos(star_name, plx, dplx, verbose)
        if d is not None:
            used_hipparcos = True
            return round(d,5), round(dd,5)

    if not used_hipparcos:
        try:
            simbad_result = Simbad.query_objectids(star_name)
            if simbad_result is None:
                if verbose:
                    print(f"No results found in Simbad for {star_name}")
                    d, dd = use_hipparcos(star_name, plx, dplx, verbose)
                return round(d, 5), round(d, dd)

            # Find the column name for 'id' (case-insensitive)
            id_col = next((col for col in simbad_result.colnames if col.lower() == "id"), None)

            if id_col is not None:
                ids = simbad_result[id_col]
            else:
                raise KeyError("No column named 'id' or 'ID' found in simbad_result")

            gaiadr3mask = ['Gaia DR3' in name for name in ids]
            if any(gaiadr3mask):
                gaiadr3_name = ids[gaiadr3mask][0]
                if verbose:
                    print("Found Gaia DR3:", gaiadr3_name)
                d, dd = gaia_correct_distance(gaiadr3_name, plx, dplx, verbose)
                return round(d,5), round(dd,5)

                # Only check DR2 if no DR3 found
            gaiadr2mask = ['Gaia DR2' in name for name in ids]
            if any(gaiadr2mask):
                gaiadr2_name = ids[gaiadr2mask][0]
                if verbose:
                    print("Found Gaia DR2:", gaiadr2_name)
                d, dd = gaia_correct_distance(gaiadr2_name, plx, dplx, verbose)
                return round(d,5), round(dd,5)

            # Only check DR1 if no DR2 found
            gaiadr1mask = ['Gaia DR1' in name for name in ids]
            if any(gaiadr1mask):
                gaiadr1_name = ids[gaiadr1mask][0]
                if verbose:
                    print("Found Gaia DR1:", gaiadr1_name)
                d, dd = gaia_correct_distance(gaiadr1_name, plx, dplx, verbose)
                return round(d,5), round(dd,5)
        except Exception as e:
            if verbose:
                print(f"Error in Simbad query: {e}")

        # Finally fall back to Hipparcos
        if verbose:
            print("No Gaia name found. Checking Hipparcos.")
        d, dd = use_hipparcos(star_name, plx, dplx, verbose)
        return round(d,5), round(dd,5)

        if d is None:
            if verbose:
                print('No valid distance could be found from Gaia or Hipparcos')
            return None, None

def calc_star_params(stellar_params, verbose=False):
    #################################################################
    # Function: calc_star_params                                    #
    # Inputs: stellar_params -> star object                         #
    #         verbose -> if True, allows the print statements       #
    # Outputs: None                                                 #
    # What it does:                                                 #
    #       1. Unpacks the star object                              #
    #       2. Calculated the physical radius and error using       #
    #          and_to_lin                                           #
    #       3. Calculates the effective temperature and error       #
    #          using temp                                           #
    #       4. Calculates the luminosity and error using luminosity #
    #       5. Updates the stellar object with the new params       #
    #################################################################
    theta = stellar_params.ldtheta
    dtheta = stellar_params.ldtheta_err
    d = stellar_params.dist
    dd = stellar_params.dist_err
    fbol = stellar_params.fbol
    dfbol = stellar_params.fbol_err

    r, dr = ang_to_lin(theta, dtheta, d, dd)
    t, dt = temp(fbol, dfbol, theta, dtheta)
    l, dl = luminosity(d, dd, fbol * (1e-8), dfbol * (1e-8))

    stellar_params.teff = round(t, 5)
    stellar_params.teff_err = round(dt,5)
    stellar_params.rad = round(r,5)
    stellar_params.rad_err = round(dr,5)
    stellar_params.lum = round(l,5)
    stellar_params.lum_err = round(dl,5)

    if verbose:
        print("Linear Radius: ", round(r, 3), "+/-", round(dr, 3), "[R_solar]")
        print("Luminosity: ", round(l, 3), "+/-", round(dl, 3), "[L_solar]")
        print("Effective temperature: ", round(t, 3), "+/-", round(dt, 3), "[K]")


class StellarParams:
    # Attributes can include:
    # fbol and fbol_err: bolometric flux and its error in units of 10^-8 ergs/ s*cm^2
    # logg and logg_err: surface gravity in units of dex
    # feh and feh_err: metallicity of star in units of dex
    # dist and dist_err: distance and its error in units of parsecs
    # plx and plx_err: parallax and its error in units of mas
    # initial udtheta and udtheta_err: uniform disk diameter and its error in units of mas
    # initial ldtheta and ldtheta_err: limb darkened disk diameter and its error in units of mas
    # final udtheta and udtheta_err: uniform disk diameter and its error in units of mas
    # final ldtheta and ldtheta_err: limb darkened disk diameter and its error in units of mas
    # teff and teff_err: temperature and its error in units of Kelvin
    # lum and lum_err: luminosity and its error in units of Lsol
    # rad and rad_err: physical radius and its error in units of Rsol

    __units__ = {
        'fbol': '10⁻⁸ erg/s/cm²',
        'fbol_err': '10⁻⁸ erg/s/cm²',
        'logg': 'dex',
        'logg_err': 'dex',
        'feh': 'dex',
        'feh_err': 'dex',
        'plx': 'mas',
        'plx_err': 'mas',
        'dist': 'pc',
        'dist_err': 'pc',
        'udthetai': 'mas',
        'udthetai_err': 'mas',
        'ldthetai': 'mas',
        'ldthetai_err': 'mas',
        'udtheta': 'mas',
        'udtheta_err': 'mas',
        'ldtheta': 'mas',
        'ldtheta_err': 'mas',
        'teff': 'K',
        'teff_err': 'K',
        'lum': 'L☉',
        'lum_err': 'L☉',
        'rad': 'R☉',
        'rad_err': 'R☉',
        'ldc_R': ' ',
        'ldc_K': ' ',
        'ldc_H': ' ',
        'ldc_J': ' '
    }

    def __init__(self, **kwargs):
        # Set all known attributes to None by default
        for key in self.__class__.__units__:
            setattr(self, key, None)
        # Then overwrite with provided values
        for k, v in kwargs.items():
            setattr(self, k, v)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        lines = []
        for k, v in self.__dict__.items():
            unit = self.__units__.get(k, '')
            unit_str = f" [{unit}]" if unit else ''
            lines.append(f"{k} = {v}{unit_str}")
        return "\n".join(lines)



import numpy as np
import pandas as pd
from lmfit import Model
import concurrent.futures
import scipy.special as ss
from radpy.stellar import temp
from astropy.stats import mad_std
from radpy.UDfitting import chis, weight_avg, percent_diff, safe_theta_extraction
from radpy.limbdarkcoeffs import ldc_calc

import warnings
warnings.filterwarnings("ignore", message="Using UFloat objects with std_dev==0 may give unexpected results.")
warnings.filterwarnings("ignore", message="DataFrameGroupBy.apply operated on the grouping columns")
#Limb-darkened disk V2 equation
def V2(sf, theta, mu):
    #mu = 0.7039725
    alpha = 1-mu
    beta = mu
    x = np.pi*sf*(theta/(206265*1000))
    vis = (((alpha/2)+(beta/3))**(-2))*((alpha*(ss.jv(1,x)/x))+ beta*(np.sqrt(np.pi/2)*(ss.jv(3/2,x)/(x**(3/2)))))**2
    return vis
##########################################################################################
# Random bracket function for bootstrapping for limb-darkening
def random_bracket_ld(df, num_of_brackets):
    ###########################################################################
    # Function: random_bracket                                                #
    # Inputs: df -> dataframe with spatial frequency, v2, v2_err, and bracket #
    #         num_of_brackets -> number of brackets in total you have         #
    # Outputs: spf_br -> the spatial frequencies randomized                   #
    #          v2_br -> the visibility squared randomized                     #
    #          dv2_br -> the error on the v2 randomized                       #
    #          ldc_br -> the limb-darkening coefficients
    #          wavgs -> weighted averages of the v2                           #
    # What it does:                                                           #
    #      1. sets the seed                                                   #
    #      2. picks a random number between 2 and the number of brackets      #
    #      3. Selects that many unique bracket labels at random               #
    #      4. Filters the dataframe down to only those with those specific    #
    #         bracket labels                                                  #
    #      5. Groups by the bracket labels                                    #
    #      6. Applies the weight average to the grouped data                  #
    #      7. Adds the weighted average as a column in the data frame         #
    #      8. Merges the groups into a new group for weighted average         #
    #         based on the bracket                                            #
    #      9. Splits up the dataframe into spatial frequency, v2, dv2, ldcs,  #
    #         and wavg                                                        #
    #     10. Returns spatial frequency, v2, dv2, and wavg                    #
    ###########################################################################
    np.random.seed()
    xdata = []
    ydata = []
    dydata = []
    ldcdata = []
    numbr = np.random.randint(2, num_of_brackets)
    # chatgpt wrote the next couple lines
    random_group_ids = df['Bracket'].drop_duplicates().sample(n=numbr).values

    # Filter the DataFrame for the selected groups
    random_groups = df[df['Bracket'].isin(random_group_ids)]

    grouped = random_groups.groupby('Bracket')
    results = grouped.apply(weight_avg).reset_index()
    results.columns = ['Bracket', 'Wavg']
    random_groups_with_avg = random_groups.merge(results, on='Bracket')

    spf_br = random_groups_with_avg['Spf']  # spatial frequency in rad^-1
    v2_br = random_groups_with_avg['V2']  # Visibility squared
    dv2_br = random_groups_with_avg['dV2']
    ldc_br = random_groups_with_avg['LDC']
    wavgs = random_groups_with_avg['Wavg']

    return spf_br, v2_br, dv2_br, ldc_br, wavgs

##########################################################################################
def initial_LDfit(spf, v2, dv2, star_params, filt, verbose=False):
    #####################################################################
    # Function: initial_LDfit                                           #
    # Inputs: spf -> spatial frequency                                  #
    #         v2 -> visibilitity squared                                #
    #         dv2 -> error on the V2                                    #
    #         theta_guess -> initial guess for theta                    #
    #         star_params -> stellar class object                       #
    #         verbose -> if set to True, allows print statements        #
    #                    defaults to False                              #
    # Outputs: ldtheta_ilm -> initial uniform disk diameter             #
    #          lddtheta_ilm -> error on the diameter                    #
    #          chisqr_ldilm -> chi squared reduced value                #
    # What it does:                                                     #
    #        1. Calculates the temperature using the initial UD         #
    #        2. Calcualted the LDC                                      #
    #        3. Initialized the model                                   #
    #        4. initializes the parameters                              #
    #        5. Fits for the UD diameter using lmfit                    #
    #           uses for the weights as 1/dv2                           #
    #        6. pulls out the theta, dtheta, and chi squared reduced    #
    #        7. updates the stellar object                              #
    #        8. Returns the theta, dtheta, and chi squared reduced      #
    #####################################################################
    t, dt = temp(star_params.fbol, star_params.fbol_err, star_params.udthetai, star_params.udthetai_err)
    ldc = ldc_calc(t, star_params.logg, star_params.feh, filt)

    ldmodel = Model(V2, independent_vars=['sf', 'mu'])
    ldparams = ldmodel.make_params(theta=star_params.udthetai)
    ld_result = ldmodel.fit(v2, ldparams, sf=spf, mu=ldc, weights= 1 / (dv2), scale_covar=False)
    ldtheta_ilm, lddtheta_ilm = safe_theta_extraction(ld_result)
    chisqr_ldilm = ld_result.redchi  # chi squared reduced of the fit

    star_params.update(ldthetai=round(ldtheta_ilm,5), ldthetai_err = round(lddtheta_ilm,5), teff=round(t,5), teff_err=round(dt,5))
    if verbose:
        print("Effective temperature:", round(t,5), "+/-", round(dt,5), "K")
        print("LDC for filter ", filt, ":", round(ldc,5))
        print('Initial fit with lmfit:')
        print(ld_result.fit_report())

    return ldtheta_ilm, lddtheta_ilm, chisqr_ldilm


def bootstrap_ld(df, inst):
    ###########################################################
    # Function: bootstrap_ld                                  #
    # Inputs: df -> the data dataframe                        #
    #         inst -> the intstrument                         #
    # Outputs: the new_df                                     #
    # What it does:                                           #
    #       1. If the instrument is set to c (Classic),       #
    #          samples the V2 on a normal distribution        #
    #          and creates a new dataframe with that.         #
    #       2. If the instrument is any others, determines    #
    #          the number of brackets in the dataset.         #
    #       3. calls the random_bracket function              #
    #       4. samples the V2 on a normal distribution        #
    #       5. creates a new dataframe with that              #
    #       6. Returns the new dataframe                      #
    ###########################################################
    if inst == 'c' or inst == 'C':
        newv2 = np.random.normal(df['V2'], df['dV2'])
        new_df = pd.DataFrame(np.column_stack((df['Spf'], newv2, df['dV2'], df['LDC'])),
                              columns=['Spf', 'V2', 'dV2', 'LDC'])
        return new_df
    else:
        num_brackets = df['Bracket'].max()
        spfbr, v2br, dv2br, ldcbr, avgdv2 = random_bracket_ld(df, num_brackets)
        newv2 = np.random.normal(v2br, avgdv2)
        new_df = pd.DataFrame(np.column_stack((spfbr, newv2, dv2br, ldcbr)), columns=['Spf', 'V2', 'dV2', 'LDC'])
        return new_df


def ldfit(df, stellar_params, verbose=False):
    #####################################################################
    # Function: ldfit                                                   #
    # Inputs: df -> dataframe with data in it                           #
    #         star_params -> stellar class object                       #
    #         verbose -> if set to True, allows print statements        #
    #                    defaults to False                              #
    # Outputs: theta_ld -> initial uniform disk diameter                #
    # What it does:                                                     #
    #        1. Initialized the model                                   #
    #        2. initializes the parameters                              #
    #        3. Fits for the LD diameter using lmfit                    #
    #           uses for the weights as 1/dv2                           #
    #        4. pulls out the theta                                     #
    #        5. Returns the theta                                       #
    #####################################################################
    ldmodel = Model(V2, independent_vars=['sf', 'mu'])
    ld_params = ldmodel.make_params(theta=stellar_params.udtheta)
    ld_result = ldmodel.fit(df['V2'], ld_params, sf=df['Spf'], mu=df['LDC'], weights=1 / (df['dV2']), scale_covar=True)
    theta_ld, _ = safe_theta_extraction(ld_result)
    #theta_ld = ld_result.uvars['theta'].n

    return theta_ld


def ldfit_values(x, y, dy, LD, ldcs, stellar_params, verbose=False):
    ##################################################################
    # Function: ldfit_values                                         #
    # Inputs: x -> the spatial frequencies                           #
    #         y -> the V2                                            #
    #        dy -> the error on the V2                               #
    #        LD -> the list of diameters                             #
    #        ldcs -> limb darkening coefficients                     #
    #        stellar_params -> the star object                       #
    #        verbose - > if true, returns print statements           #
    # Outputs: avg_LD -> average limb darkened diameter              #
    #          std_LD -> the median absolute deviation of LD theta   #
    #          teff_ld[0] -> effective temperature                   #
    #          teff_ld[1] -> error on the effective temperature      #
    #          ldc_results -> the ldc for each band                  #
    #          chisq_results -> the chi square and chi square red    #
    #                           values for each ldc band             #
    # What it does:                                                  #
    #     1. Takes the mean of the limb-darkened disk diameters      #
    #     2. Takes the median absolute deviation of the LDs          #
    #     3. Calculates the effective temperature using the mean     #
    #     4. Initializes the ldc_results and chisq_results           #
    #        to store dynamically                                    #
    #     5. For each band in the ldcs, calculates the V2 model      #
    #     6. Calculates the chi squared and chi squared reduced      #
    #        for each LDC band                                       #
    #     7. Stores the results in the ldc_results and chisq_results #
    #     8. Returns the avg_LD, std_LD, teff and teff error, the    #
    #        ldc results, and the chi squared results                #
    ##################################################################
    avg_LD = np.mean(LD)
    std_LD = mad_std(LD)
    teff_ld = temp(stellar_params.fbol, stellar_params.fbol_err, avg_LD, std_LD)
    # Store results dynamically
    ldc_results = {}
    chisq_results = {}

    for band in ldcs:
        ldc_val = ldcs[band]
        if ldc_val is not None:
            model_v2 = V2(x, avg_LD, ldc_val)
            chisq, chisqr = chis(y, model_v2, dy, 1)
            ldc_results[band] = ldc_val
            chisq_results[band] = {"chisq": chisq, "chisqr": chisqr}

    if verbose:
        print('Limb-darkened Disk Diameter after MC/BS:', round(avg_LD, 4), '+/-', round(std_LD, 5), 'mas')
        for band, ldc_val in ldc_results.items():
            print(f"Limb-darkening coefficient in {band}:", round(ldc_val, 5))
            print(f"Chi-squared for {band} band:", round(chisq_results[band]["chisq"], 3))
            print(f"Reduced chi-squared for {band} band:", round(chisq_results[band]["chisqr"], 3))
        print("Temperature:", round(teff_ld[0], 1), "+/-", round(teff_ld[1], 1), "K")

    return avg_LD, std_LD, teff_ld[0], teff_ld[1], ldc_results, chisq_results


def mcbs_worker(args):
    #############################################################
    # Function: mcbs_worker                                     #
    # Inputs: args -> the mc_dfs, the bs_num, stellar_params,   #
    #                 and verbose                               #
    # Outputs: the limb darkened disk list                      #
    # What it does:                                             #
    #       1. unpacks the arguments                            #
    #       2. Initializes the limb-darkened disk list          #
    #       3. Enters the bootstrap loop                        #
    #       4. For each dataframe created in the Monte Carlo    #
    #          loop, it determines which instrument, then       #
    #          calls the bootstrap function for the ld          #
    #       5. Appends the resulting dataframe to the list      #
    #       6. Concatenates all the bootstrapped dfs into one   #
    #       7. Calls ldfit and fits for the limb-darkened theta #
    #       8. Appends results to the LD list                   #
    #       9. Returns the LD list                              #
    #############################################################

    mc_dfs, bs_num, stellar_params, verbose = args
    LD = []
    for _ in range(bs_num):
        bs_dfs = []
        for df in mc_dfs:
            inst = df["Instrument"].iloc[0]
            boot_df = bootstrap_ld(df, inst)
            bs_dfs.append(boot_df)
        new_df = pd.concat(bs_dfs, ignore_index=True)
        theta_ldbs = ldfit(new_df, stellar_params, verbose)
        LD.append(theta_ldbs)
    return LD


def run_LDfit(mc_num, bs_num, ogdata, datasets, stellar_params, verbose=False, debug=False):
    ######################################################################
    # Function: run_ldmcbs_fit_parallel                                  #
    # Inputs: mc_num -> number of Monte Carlo iterations                 #
    #         bs_num -> number of bootstrap iterations                   #
    #         ogdata -> original data sets                               #
    #         datasets -> the datasets you want fit                      #
    #                     format: [inst1, inst2, inst3]                  #
    #         stellar_params -> star object                              #
    #         verbose -> if True, allows print statements                #
    #                    default is False                                #
    #         debug -> allows debug statements to show                   #
    #                  default is set to False                           #
    # Outputs: theta_ld-> final limb-darkened disk diameter              #
    #          dtheta_ld -> error on the ld diameter                     #
    #          T -> effective temperature                                #
    #          dT -> error on the effective temperature                  #
    #          final_ldcs -> the final ldcs for each detected band       #
    #          final_chisqrs -> the final chi square and chi square      #
    #                           reduced values                           #
    # What it does:                                                      #
    #      1. Initializes a filter map dictionary relating each          #
    #         instrument to a filter                                     #
    #      2. sets the T_new to be the current temperature in the star   #
    #         object                                                     #
    #      3. sets an arbitrary number for the diff_teff and diff_theta  #
    #      4. sets the minimum percent difference                        #
    #      5. unpacks the ogdata for comparison later                    #
    #      6. Starts the while loop that compares the percent difference #
    #         between the theta and teff of the iteration before and the #
    #         theta and teff of the current iteration                    #
    #      7. Initializes the empty list for the diameters and a         #
    #         dynamic list for the ldcs per filter                       #
    #      8. For each data set, it calculates a ldc depending on the    #
    #         instrument                                                 #
    #      9. enters the Monte Carlo loop                                #
    #     10. Creates the dataframes for each dataset                    #
    #     11. For each dataframe, it samples the ldc on a normal         #
    #         distribution.
    #     12. For each dataframe, samples the wavelength of observation  #
    #         on a normal distribution. Then calculates new spatial      #
    #         frequencies                                                #
    #     13. Begins running all the bootstrapping loops in parallel     #
    #         by calling mcbs_worker                                     #
    #     14. Appends each result of the mcbs_worked to the LD list      #
    #     15. Resets the while loop iterators                            #
    #     16. Calcualtes a new LD theta, LD dtheta, teff and dteff by    #
    #         calling ldfit_values                                       #
    #     17. Updates the stellar object                                 #
    #     18. Calculates the new percent difference for theta and teff   #
    #     19. After the final iteration of the while loop,               #
    #         calls ldfit_values to do a final fit for the theta, theta  #
    #         error, teff, teff error, ldc_values, and chi-square vals   #
    #     20. Updates stellar object with the ldc values for each filter #
    #     21. Calculates final percent differences for teff and theta    #
    #     22. Returns final theta, theta err, teff, teff error, ldc_vals #
    #         and chi-sqr vals.                                          #
    ######################################################################
    filter_map_i = {
        'p': 'R',
        'v': 'R',
        'c': 'K',
        'm': 'H',
        'my': 'K',
        's': 'R'
        # Add other instruments as needed
    }
    T_new = stellar_params.teff
    theta_new = stellar_params.udtheta
    diff_theta = 5
    diff_teff = 5
    min_percent = 0.05
    iter = 0
    x = ogdata[0]
    y = ogdata[1]
    dy = ogdata[2]
    while diff_theta >= min_percent or diff_teff >= min_percent:
        LD = []
        ldc_per_filter = {}
        for d in datasets:
            inst = d.instrument.lower()
            filt = filter_map_i[inst]
            if filt not in ldc_per_filter:
                ldc_val = ldc_calc(stellar_params.teff,
                                   stellar_params.logg,
                                   stellar_params.feh, filt)
                ldc_per_filter[filt] = ldc_val

        mc_args = []
        for _ in range(mc_num):
            mc_dfs = []
            for d in datasets:
                inst = d.instrument.lower()
                filt = filter_map_i[inst]
                mu = np.random.normal(ldc_per_filter[filt], 0.02)
                df = d.make_df(LDC=mu)
                df['Spf'] = df['B'] / np.random.normal(df['Wave'], df['Band'])
                mc_dfs.append(df)
            mc_args.append((mc_dfs, bs_num, stellar_params, verbose))

        # Parallel execute
        with concurrent.futures.ThreadPoolExecutor() as executor:
            results = list(executor.map(mcbs_worker, mc_args))
            for res in results:
                LD.extend(res)

        T_old = T_new
        theta_old = theta_new
        theta_new, _, T_new, _, _, _ = ldfit_values(x, y, dy, LD, ldc_per_filter, stellar_params, verbose=debug)
        stellar_params.update(teff=round(T_new,5), ldtheta=round(theta_new,5))
        diff_teff = percent_diff(T_old, T_new, verbose=debug)
        diff_theta = percent_diff(theta_old, theta_new, verbose=debug)
        iter += 1
    if verbose:
        print("Final Values after ", iter, " iterations:")
    theta_ld, dtheta_ld, T, dT, final_ldcs, final_chis = ldfit_values(x, y, dy, LD, ldc_per_filter, stellar_params,
                                                                      verbose)
    stellar_params.update(teff=round(T,5), ldtheta=round(theta_ld,5), ldtheta_err=round(dtheta_ld,5))
    for filt, mu in final_ldcs.items():
        setattr(stellar_params, f"ldc_{filt}", round(mu, 5))
    diff_teff = percent_diff(T_old, T_new, verbose)
    diff_theta = percent_diff(theta_old, theta_new, verbose)

    return theta_ld, dtheta_ld, T, dT, final_ldcs, final_chis




import os

raddir = os.path.dirname(os.path.realpath(__file__))
datadir = os.path.join(raddir, 'data')

hapath = os.path.join(datadir, 'ldcHA.csv')
hppath = os.path.join(datadir, 'ldcHP.csv')
japath = os.path.join(datadir, 'ldcJA.csv')
jppath = os.path.join(datadir, 'ldcJP.csv')
kapath = os.path.join(datadir, 'ldcKA.csv')
kppath = os.path.join(datadir, 'ldcKP.csv')
rapath = os.path.join(datadir, 'ldcRA.csv')
rppath = os.path.join(datadir, 'ldcRP.csv')

classicpath = os.path.join(datadir, 'ClassicData.csv')
pavopath = os.path.join(datadir, 'PAVOdata.csv')
vegapath = os.path.join(datadir, 'Vegadata.csv')



import re
import os
import pandas as pd
from radpy.stellar import *
from radpy.datareadandformat import *
from radpy.plotting import plot_v2_fit
from radpy.LDfitting import initial_LDfit, run_LDfit
from radpy.UDfitting import initial_UDfit, run_UDfit, udfit_values

def extract_id(star_name):
    match = re.search(r'\d+', star_name)
    return match.group(0) if match else None


def find_files_for_star(star_id, data_dir):
    matches = []
    for root, dirs, files in os.walk(data_dir):
        for file in files:
            if star_id in file:
                matches.append(os.path.join(root, file))
    return matches

def extract_instrument_from_filename(filename):
    instruments = {'C': 'classic', 'P': 'pavo', 'V': 'vega', 'M': 'mircx', 'MY': 'mystic', 'S': 'spica'}
    name = os.path.basename(filename).upper()
    # Try to detect instrument by filename
    if "PAVO" in name: return "P"
    if "VEGA" in name: return "V"
    if "CLASSIC" in name or "CLIMB" in name: return "C"
    if "MIRCX" in name: return "M"
    if "MYSTIC" in name: return "MY"
    if "SPICA" in name: return "S"
    # Fallback: use first letter prefix
    for key in instruments:
        if name.startswith(key):
            return key
    return "UNKNOWN"

def save_plot(fig, out_dir, star_id, plot_type, extension):
    os.makedirs(out_dir, exist_ok=True)
    if extension == '.png':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.png"), bbox_inches = 'tight', dpi = 200)
    if extension == '.jpg':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.jpg"), bbox_inches='tight', dpi = 200)
    if extension == '.eps':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.eps"), bbox_inches='tight', dpi = 200)
    if extension == '.pdf':
        fig.savefig(os.path.join(out_dir, f"{star_id}_{plot_type}.pdf"), bbox_inches='tight', dpi = 200)

def write_latex_table(df, out_file):
    with open(out_file, "w") as f:
        f.write(df.to_latex(index=False, float_format="%.4f"))

def get_stellar_params(csv_path):
    """
    Read stellar parameters for each star from a CSV with columns:
    star_name, fbol, fbol_err, logg, logg_err, feh, feh_err
    """
    df = pd.read_csv(csv_path, sep='\t')
    star_names = df['Star'].tolist()
    params_dict = {}
    for _, row in df.iterrows():
        params_dict[row['Star']] = {
            'fbol': row['fbol'],
            'fbol_err': row['dfbol'],
            'logg': row['logg'],
            'logg_err': row['dlogg'],
            'feh': row['feh'],
            'feh_err': row['dfeh'],
        }
    return star_names, params_dict


def converttoIDobj(dfs, verbose=False):
    # Allows for dynamic wrapping of the different instrument classes
    instrument_class_map = {
        "P": PavoData,
        "V": VegaData,
        "C": ClassicData,
        "M": MircxData,
        # "My": MysticData,
        # "S": SpicaData
    }
    wrapped_data = {}
    for inst_code, df in dfs.items():
        data_class = instrument_class_map.get(inst_code)
        if data_class:
            try:
                wrapped_data[inst_code] = data_class(df)
            except Exception as e:
                print(f"Failed to wrap data for {inst_code}: {e}")
        else:
            print(f"No class found for '{inst_code}'")
    return wrapped_data


def data_dict_plotting(wrapped_data):
    instrument_name_map = {
        "P": "pavo",
        "V": "vega",
        "C": "classic",
        "M": "mircx",
        "My": "mystic",
        "S": "spica"
    }

    data_dict = {
        instrument_name_map[code]: data
        for code, data in wrapped_data.items()
        if code in instrument_name_map
    }

    return data_dict

def format_catalog_name(name):
    # Step 1: Replace underscores with spaces
    if name.startswith('$\\rm') and name.endswith('$'):
        return name
    name = name.replace('_', ' ')
    # Step 2: Insert a space between the prefix and the numbers—only at the first transition
    name = re.sub(r'^([A-Za-z+\-]+)\s*([0-9].*)', r'\1 \2', name)
    # Step 3: Replace spaces with LaTeX non-breaking spaces
    return rf'$\rm {name.replace(" ", "~")}$'

def convert_names_to_latex(names):
    names2 = [format_catalog_name(name) for name in names]
    return names2


def process_star(star_name, data_dir, output_dir, stellar_param_dict, latex_rows, mc_num=71, bs_num=71,
                 set_axis=None, image_ext=None, binned=None, ldc_band=None, verbose=True):
    star_id = extract_id(star_name)
    print("--------------------------------------------------")
    print(f"Starting processing for {star_name}")
    files = find_files_for_star(star_id, data_dir)
    if not files:
        print(f"No files found for {star_name} ({star_id})")
        return

    # Group files by instrument
    grouped_files = {}
    for file in files:
        inst = extract_instrument_from_filename(file)
        grouped_files.setdefault(inst, []).append(file)

    # Process files by instrument
    dataframes = {}
    for inst, filelist in grouped_files.items():
        dfs = []
        for file in filelist:
            try:
                df, _ = filename_extension(file, inst, verbose=verbose)
                dfs.append(df)
            except Exception as e:
                print(f"Failed to process {file} ({inst}): {e}")
        if dfs:
            # Concatenate all files for this instrument
            dataframes[inst] = pd.concat(dfs, ignore_index=True)

    # Convert to InterferometryData objects and wrap into instrument classes
    wrap_data = converttoIDobj(dataframes, verbose=False)

    # Combine all available data using RADPy's combined() utility
    combined_args = [obj.make_df() for obj in wrap_data.values()]

    b, v2, dv2, wave, band, brack, inst = combined(*combined_args)
    spf = b / wave

    # Set up stellar parameters
    star = StellarParams()
    params = stellar_param_dict.get(star_name, {})
    for param, value in params.items():
        setattr(star, param, value)
    # Calculate distance
    if params.get('plx', None) is not None:
        D, dD = distances(star_name, plx=params['plx'], dplx=params['plx_err'], verbose=verbose)
        star.dist = D
        star.dist_err = dD
    else:
        D, dD = distances(star_name, verbose=verbose)
        star.dist = D
        star.dist_err = dD

    # Initial fits
    initial_UDfit(spf, v2, dv2, 0.4, star, verbose=verbose)
    initial_LDfit(spf, v2, dv2, star, 'R', verbose=verbose)

    # Monte Carlo uniform-disk fit
    datasets = list(wrap_data.values())
    results_UD = run_UDfit(bs_num, mc_num, datasets=datasets, stellar_params=star)
    udfit_values(spf, v2, dv2, results_UD, stellar_params=star, verbose=verbose)

    # Monte Carlo limb-darkened fit
    run_LDfit(bs_num, mc_num, ogdata=[spf, v2, dv2], datasets=datasets, stellar_params=star, verbose=verbose)

    # Calculate additional stellar parameters
    calc_star_params(star, verbose=verbose)

    # Save plots
    plot_dir = os.path.join(output_dir, "plots")
    os.makedirs(plot_dir, exist_ok=True)

    data_dict = data_dict_plotting(wrap_data)
    if binned:
        bin_only = [k for k in binned if k in data_dict]
    else:
        bin_only = None

    star_title = convert_names_to_latex([star_name])
    # Uniform disk plot
    fig1, _ = plot_v2_fit(
        data_dict=data_dict,
        star=star,
        datasets_to_plot=list(data_dict.keys()),
        to_bin=bin_only,
        plot_udmodel=True,
        eq_text=True,
        set_axis=set_axis,
        title=star_title[0],
        show=False
    )
    save_plot(fig1, plot_dir, star_id, "UDfit", image_ext)

    # Limb-darkened R band plot
    fig2, _ = plot_v2_fit(
        data_dict=data_dict,
        star=star,
        datasets_to_plot=list(data_dict.keys()),
        to_bin=bin_only,
        plot_ldmodel=True,
        ldc_band=ldc_band,
        title=star_title[0],
        set_axis=set_axis,
        eq_text=True,
        show=False
    )
    save_plot(fig2, plot_dir, star_id, "LDfit", image_ext)

    # Collect results for LaTeX
    latex_rows.append({
        "Star": star_name,
        "D (pc)": star.dist,
        r"$\Delta \rm D (pc)$": star.dist_err,
        r"$\theta_{\rm UD}$ (mas)": star.udtheta,
        r"$\Delta\theta_{\rm UD}$ (mas)": star.udtheta_err,
        r"$\theta_{\rm LD}$ (mas)": star.ldtheta,
        r"$\Delta\theta_{\rm LD}$ (mas)": star.ldtheta_err,
        r"$T_{\rm eff}$ (K)": star.teff,
        r"$\Delta T_{\rm eff}$ (K)": star.teff_err,
        r"$L_{\star} (\rm L_{\odot})$": star.lum,
        r"$\Delta L_{\star} (\rm L_{\odot})$": star.lum_err,
        r"$R_{\star} (\rm R_{\odot})$": star.rad,
        r"$\Delta R_{\star} (\rm R_{\odot})$": star.rad_err,
        r"$\mu_{\rm R}$": star.ldc_R,
        r"$\mu_{\rm K}$": star.ldc_K,
        r"$\mu_{\rm H}$": star.ldc_H,
        r"$\mu_{\rm J}$": star.ldc_J,
    })

    print(f"Finished processing {star_name}")

def batch_mode(star_file, data_dir, output_dir, latex_out, mc_num=71, bs_num=71, set_axis = None, image_ext=None, binned=None, ldc_band=None, verbose=True):
    os.chdir(data_dir)
    star_names, star_params = get_stellar_params(star_file)
    latex_rows = []
    count = 0
    for star_name in star_names:
        process_star(star_name, data_dir, output_dir, star_params, latex_rows, mc_num=mc_num, bs_num=bs_num, set_axis = set_axis,
                     image_ext=image_ext, binned=binned, ldc_band=ldc_band, verbose=verbose)
        count += 1

    latex_df = pd.DataFrame(latex_rows)
    write_latex_table(latex_df, latex_out)
    print(f"Batch complete. Fit {count} stars. Plots in {os.path.join(output_dir, 'plots')}, results in {latex_out}")